﻿using System.Collections.Generic;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.ProgramSynthesis {
  /// <summary>
  /// Replaces all occurrences of the second lit-type item with the first lit-type item in the top type vector.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  [StorableClass]
  public abstract class VectorReplaceExpression<T> : StatelessExpression {
    protected VectorReplaceExpression() { }
    [StorableConstructor]
    protected VectorReplaceExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IPushStack<IReadOnlyList<T>> vectorStack, IPushStack<T> literalStack) {
      return vectorStack.IsEmpty || literalStack.Count < 2;
    }

    protected void Eval(IPushStack<IReadOnlyList<T>> vectorStack, IPushStack<T> literalStack) {
      var vector = vectorStack.Top;
      var first = literalStack.Top;
      var second = literalStack[1];
      literalStack.Remove(2);

      var result = new List<T>(vector);
      for (var i = 0; i < vector.Count; i++) {
        if (vector[i].Equals(second))
          result[i] = first;
      }

      vectorStack.Top = result;
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.IntegerVector,
    "INTEGER[].REPLACE",
    "Replaces all occurrences of the second INTEGER with the INTEGER in the top INTEGER[].",
    StackTypes.Integer)]
  public class IntegerVectorReplaceExpression : VectorReplaceExpression<long> {
    public IntegerVectorReplaceExpression() { }
    [StorableConstructor]
    protected IntegerVectorReplaceExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.FloatVector,
    "FLOAT[].REPLACE",
    "Replaces all occurrences of the second FLOAT with the FLOAT in the top FLOAT[].",
    StackTypes.Float)]
  public class FloatVectorReplaceExpression : VectorReplaceExpression<double> {
    public FloatVectorReplaceExpression() { }
    [StorableConstructor]
    protected FloatVectorReplaceExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.FloatVectorStack, interpreter.FloatStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatVectorStack, interpreter.FloatStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.BooleanVector,
    "BOOLEAN[].REPLACE",
    "Replaces all occurrences of the second BOOLEAN with the BOOLEAN in the top BOOLEAN[].",
    StackTypes.Boolean)]
  public class BooleanVectorReplaceExpression : VectorReplaceExpression<bool> {
    public BooleanVectorReplaceExpression() { }
    [StorableConstructor]
    protected BooleanVectorReplaceExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.StringVector,
    "STRING[].REPLACE",
    "Replaces all occurrences of the second STRING with the STRING in the top STRING[].",
    StackTypes.String)]
  public class StringVectorReplaceExpression : VectorReplaceExpression<string> {
    public StringVectorReplaceExpression() { }
    [StorableConstructor]
    protected StringVectorReplaceExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.StringVectorStack, interpreter.StringStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringVectorStack, interpreter.StringStack);
    }
  }
}
